# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_core.ipynb.

# %% auto 0
__all__ = ['any2set', 'synset']

# %% ../nbs/00_core.ipynb 4
from typing import Iterable, Optional
from atyp import (
    AnyQ, BoolQ, FloatQ, StrLikeQ,    
    ClsStrQ, ClsBool, ClsFloat,
)

from indoc import indoc
from fuzstr import fuzstr
from nymstr import nymstr
from ispec.types import isiter, isset
from ispec.utils import typenone

# %% ../nbs/00_core.ipynb 6
def any2set(obj: AnyQ, hush: bool = True) -> set:
    '''Tries to convert any object to a set. If the object is a string, it will be converted to a set of length 1.'''
    if isinstance(obj, str):        
        obj = set((obj, ))

    elif not isiter(obj):
        # not a string and not iterable
        obj = set((obj, ))

    elif isinstance(obj, (list, tuple, Iterable)) and not isset(obj, set):
        obj = set(obj)

    elif hasattr(obj, '__hash__'):
        obj = {obj, }

    else:
        try:
            obj = set(obj)
        except TypeError:
            if not hush:
                raise TypeError(f'Could not convert {obj} to set.')
            obj = set()

    return typenone(obj, set)

# %% ../nbs/00_core.ipynb 8
@indoc
class synset(nymstr):    
    '''synstr (Alias-able String Class) extends the nymstr type to handle synonyms.

    Parameters
    ----------
    canon : str
        The canonical name of the string.

    *synonyms : Iterable[str]
        The synonyms of the string. They will be added to the set alongside the canonical name.

    icase : bool, default: False
        Whether to ignore case when matching, by default False.

    ispace : bool, default: False
        Whether to ignore spaces when matching, by default False.

    iunder : bool, default: False
        Whether to ignore underscores when matching, by default False.
        
    cutoff : float, default: 1
        The cutoff for fuzzy matching, by default 1.

    itemattr : ClsStrQ, default: 'synonyms'
        The attribute which holds the item being named, by default 'synonyms'.

    synsattr: ClsStrQ, default: 'synonyms'
        The attribute which holds the synonyms set, by default 'synonyms'.

    fullrepr : ClsBool, default:  False
        Whether or not to show the full representation, by default False.

    fuzzsyns : ClsBool, default: True
        Whether or not to make all synonyms fuzzy, by default True.

    Attributes
    ----------
    syns : set
        The dynamic `synsattr` attribute representing the synonym set.

    alts : set
        The synonyms (canon excluded).
        
    nalts : int
        The number of synonyms (canon excluded).

    raw : str
        The raw string.

    canon : str
        The canonical string (i.e. the raw string))
    
    known : set
        The set of known strings (canon and synonyms included)

    snake : str
        The snake-cased string.

    item : set
        The item which was named. In this case it is a set of synonyms (canon included).

    aspec : list | tuple | dict, default: ('icase', 'ispace', 'iunder', 'cutoff', 'itemattr', 'synsattr', 'fullrepr', 'fuzzsyns', )
        The attribute specification. Must be either an iterable of strings 
        or a dictionary of `str: Any`. Defaults to an empty tuple.
    
    dspec : list | tuple | dict, default: ('itemattr', 'synsattr',)
        The dynamic attribute specification. Must be either an iterable of strings 
        or a dictionary of `str: str`. Defaults to an empty tuple.

    specs: {('aspec', ), ('aspec', 'dspec', ), ('aspec', 'dspec), (), }, default: (Spec.aspec.value, Spec.dspec.value)
        The attribute specifications to use.
        
    __readonly__ : tuple, default: (Spec.aspec.value, Spec.dspec.value)
        Attributes that cannot be set.

    Methods
    -------
    iscanon(other: str) -> bool:
        Check if `other` is the instance's canonical name.
        
    isknown(other: str, extra: SetQ = None) -> bool:
        Check if the fuzzified `other` matches any of the known (and extra) strings.

    issyn(other: str, extra: Optional[set] = None) -> bool:
        Check if other is a synonym of the instance.

    issyns(other: str) -> bool:
        Check if other's synonym set is equivalent.

    obj2set(other) -> set:
        Convert `other` to a set.

    isitem(other: AnyQ) -> bool:
        Check if `other` is the instance's item.

    tostr() -> str:
        Convert the instance to a string.

    prep(s: StrLikeQ = None) -> str:
        Prepare a string by applying case and space insensitivity rules.

    find(strseq: Iterable[StrLike], **kwargs) -> List[str]:
        Find the closest matches (same-ish) in the provided sequence using fuzzy matching.

    iseq(other: str) -> bool:
        Check if the provided string is an alias of this entity using fuzzy matching.

    __eq__(other):
        Check if the provided value is an alias of this entity using fuzzy matching.

    __str__():
        Return the raw string.

    __hash__():
        Return the hash of the raw string.

    __setattr__(name, value):
        Set the class attribute if it's not read-only.

    getattrkeys(spec: str, dyn: bool = False) -> tuple[str, ...]:
        Return attribute keys stored in `spec`.

    getattrvals(self, spec: str, dyn: bool = False) -> tuple[Any, ...]:
        Return default attribute values stored in `spec`.
    
    skeys() -> tuple[str, ...]:
        Return all attribute keys stored for each spec stored in `specs`.

    svals() -> tuple[str, ...]:
        Return default attribute values stored for each spec stored in `specs`.
    
    getattrs(**kwargs):
        Get instance parameters with optional overrides.
        
    makesame(*args, **kwargs):
        Call class constructor with the same attributes as the current instance.
    
    isinst(other):
        Check if the provided value is an instance of this class.
        
    sameattrs(other):
        Check if the provided value is an instance of this class with the same attributes.
        
    diffattrs(other):
        Check if the provided value is an instance of this class with different attributes.

    getattrname(dattr: str) -> str:
        Get the name of the dynamic attribute.
    
    getdattr(dattr: str, default: Any = None) -> Any:
        Get the value of the dynamic attribute.

    setdattr(dattr: str, value: Any = None):
        Set the value of the dynamic attribute.

    update_attrname(dattr: str, **kwargs):
        Update the name of the dynamic attribute.
    
    update_dattrval(aname: str, **kwargs):
        Update the value of the dynamic attribute.
    
    update_dattr(dattr: str, **kwargs):
        Update the name and value of the dynamic attribute.

    update_aspec(**kwargs):
        Update the attribute specification.

    update_dspec(**kwargs):
        Update the dynamic attribute specification.

    update_specs(**kwargs):
        Update the specifications in `specs` e.g. `aspec` and / or `dspec`.

    getclsattr(attr: str, default: Any = None) -> Any:
        Get the value of the class attribute.

    setclsattr(attr: str, val: Any = None):
        Set the value of the class attribute.

    copy():
        Return a shallow copy of the instance.
    
    deepcopy():
        Return a deep copy of the instance.

    __contains__(syn: str):
        Check if a synonym is in the set.
    
    __len__():
        Return the number of synonyms in the set (canon included).
    
    __str__():
        Return the string for the canonical name.

    __iter__():
        Return an iterator over the set.
    
    __repr__():
        Return a string representation of the object.

    __and__(other):
        Return the intersection of two sets.

    __ge__(other):
        Check if one set is a superset of another.
        
    __gt__(other):
        Check if one set is a proper superset of another.
        
    __iand__(other):
        Update a set with the intersection of itself and another.
        
    __ior__(other):
        Update a set with the union of itself and another.
        
    __isub__(other):
        Remove all elements of another set from this set.
        
    __ixor__(other):
        Update a set with the symmetric difference of itself and another.
        
    __le__(other):
        Check if one set is a subset of another.
        
    __lt__(other):
        Check if one set is a proper subset of another.
        
    __or__(other):
        Return the union of sets as a new set.
        
    __rand__(other):
        Return the intersection of two sets as a new set.
        
    __ror__(other):
        Return the union of sets as a new set.

    __xor__(other):
        Return the symmetric difference of two sets as a new set.

    __sub__(other):
        Return the difference of two sets as a new set.
        
    __rsub__(other):
        Return the difference of two or more sets as a new set.
        
    __rxor__(other):
        Return the symmetric difference of two sets as a new set.
        
    add(syn: str):
        Add a synonym to the set.
        
    remove(syn: str):
        Remove a synonym from the set.
        
    clear():
        Remove all synonyms from the set.
        
    difference(*others):
        Return the difference of two or more sets as a new set.
        
    difference_update(*others):
        Remove all elements of another set from this set.
        
    discard(syn: str):
        Remove a synonym from the set.
        
    intersection(*others):
        Return the intersection of two or more sets as a new set.
        
    intersection_update(*others):
        Update a set with the intersection of itself and another.
        
    isdisjoint(other):
        Return True if two sets have a null intersection.
        
    issubset(other):
        Report whether another set contains this set.
        
    issuperset(other):
        Report whether this set contains another set.
        
    pop():
        Remove and return an arbitrary set element.
        
    remove(syn: str):
        Remove a synonym from the set.
        
    symmetric_difference(other):
        Return the symmetric difference of two sets as a new set.
        
    symmetric_difference_update(other):
        Update a set with the symmetric difference of itself and another.
        
    union(*others):
        Return the union of sets as a new set.
        
    update(*others):
        Update a set with the union of itself and others.
        
    ''' 
    icase:  ClsBool  = True # Whether to ignore case when matching
    ispace: ClsBool  = True # Whether to ignore spaces when matching
    iunder: ClsBool  = True # Whether to ignore underscores when matching
    cutoff: ClsFloat = 0.8  # The cutoff for fuzzy matching
    
    itemattr: ClsStrQ = 'synonyms'
    synsattr: ClsStrQ = 'synonyms'
    fullrepr: ClsBool = False
    fuzzsyns: ClsBool  = True # Make all synonyms fuzstrs

    aspec = ('icase', 'ispace', 'iunder', 'cutoff', 'itemattr', 'synsattr', 'fullrepr', 'fuzzsyns', )
    dspec = ('itemattr', 'synsattr',)

    # def makecopy(self, *args, **kwargs):
    #     '''Call class constructor with the same attributes as the current instance.'''
    #     return type(self)(self, *args, **self.getattrs(**kwargs))
    
    # def __copy__(self):
    #     return super().__copy__(self, *self.syns, **self.attrs)

    def __new__(
        cls, canon: str, *synonyms, 
        icase:  BoolQ  = True,           # Case insensitive
        ispace: BoolQ  = True,           # Space insensitive
        iunder: BoolQ  = True,           # Underscore insensitive
        cutoff: FloatQ = 0.8,            # Fuzzy cutoff
        itemattr: StrLikeQ = 'synonyms', # Dynamic item attribute
        synsattr: StrLikeQ = 'synonyms', # Dynamic synonym attribute
        fullrepr: BoolQ = True,          # Full representation
        fuzzsyns: BoolQ = True,          # Make all synonyms fuzstrs
        **kwargs
    ):
        # Put all variables in a dictionary for easy access
        synonyms = {canon, *synonyms} if synonyms else {canon, }
        kwargs.update(
            icase=icase, ispace=ispace, iunder=iunder, cutoff=cutoff,
            itemattr=itemattr, synsattr=synsattr, fullrepr=fullrepr,
            fuzzsyns=fuzzsyns,
        )
        
        if fuzzsyns:
            synonyms = {fuzstr(syn, **kwargs) for syn in synonyms}

        kwargs[synsattr] = synonyms
        if kwargs.get(itemattr, None) is None:
            kwargs[itemattr] = synonyms
        
        obj = super().__new__(cls, canon, **kwargs)
        return obj
    
    def __init__(
        self, canon: str, *synonyms, 
        icase:  BoolQ   = True,          # Case insensitive
        ispace: BoolQ   = True,          # Space insensitive
        iunder: BoolQ   = True,          # Underscore insensitive
        cutoff: FloatQ  = 0.8,           # Fuzzy cutoff
        itemattr: StrLikeQ = 'synonyms', # Dynamic item attribute
        synsattr: StrLikeQ = 'synonyms', # Dynamic synonym attribute
        fullrepr: BoolQ = True,          # Full representation
        fuzzsyns: BoolQ  = True,         # Make all synonyms fuzstrs
        **kwargs
    ):
        # Put all variables in a dictionary for easy access

        synonyms = {canon, *synonyms} if synonyms else {canon, }
        kwargs.update(
            icase=icase, ispace=ispace, iunder=iunder, cutoff=cutoff,
            itemattr=itemattr, synsattr=synsattr, fullrepr=fullrepr,
        )
        if fuzzsyns:
            synonyms = {fuzstr(syn, **kwargs) for syn in synonyms}

        kwargs[synsattr] = synonyms        
        if kwargs.get(itemattr, None) is None:
            kwargs[itemattr] = synonyms

        super().__init__(canon, **kwargs)

    def __hash__(self):
        return super().__hash__()
    
    @property
    def syns(self) -> set:
        '''The dynamic `synsattr` i.e. `getattr(self, self.synsattr, None)` representing the synonym set'''
        return self.getdattr('synsattr', set())
    
    @property
    def alts(self) -> set:
        '''Synonyms (canon excluded)'''
        return (self.syns or set()) ^ set((self.canon, ))
    
    @property
    def known(self) -> set: 
        return {self.canon, *self.alts, *(self.item or set({}))}
    
    @property
    def nalts(self):
        return max(len(self.alts), 0)
    
    def issyn(self, other: str, extra: Optional[set] = None) -> bool:
        '''Check if other is a synonym of the instance.'''
        return self.isknown(other, extra)
    
    def issyns(self, other: str) -> bool:
        '''Check if other's synonym set is equivalent.'''
        sset = getattr(self,  'syns', None)
        oset = getattr(other, 'syns', None)
        return sset == oset
    
    # NOTE: we override this from nymstr
    def isitem(self, other: AnyQ) -> bool:
        '''Check if `other` is the instance's item'''        
        sitem = getattr(self,  'item', None)
        oitem = getattr(other, 'item', None)
        if isinstance(other, type(self)) and hasattr(other, 'item'):
            if isset(sitem) and isset(oitem):
                return any(oitem in self.item for oitem in other.item)
        return other in sitem

    def __eq__(self, other):
        '''Check equality based on elements in the set.'''
        if isinstance(other, (type(self))):
            return self.iscanon(other.canon) and self.issyn(other)
        return other in self

    def __contains__(self, item):
        '''Check if a synonym is in the set'''
        return self.issyn(item) 
    
    def __len__(self):
        syns = (self.known or set())
        return len(syns)

    def __str__(self):
        '''Return the canonical name.'''
        return self.canon
    
    def __iter__(self):
        '''Iterate through the set'''
        return iter(self.known)
    
    def __repr__(self):
        '''Return a string representation of the object.'''
        sub = 'alias'  if self.nalts == 1 else 'aliases'
        if not self.fullrepr:
            return super().__str__()
        return f'{self.canon} ({self.nalts} {sub})'

    def obj2set(self, other) -> set:
        if isinstance(other, type(self)) and hasattr(other, 'known') and other.known:
            return other.known        
        return any2set(other)

    def __and__(self, other):
        '''Return the intersection of two sets.'''
        return self.known & self.obj2set(other)
    def __ge__(self, other):
        '''Check if one set is a superset of another.'''
        return self.known >= self.obj2set(other)
    def __gt__(self, other):
        '''Check if one set is a proper superset of another.'''
        return self.known > self.obj2set(other)
    def __iand__(self, other):
        '''Update a set with the intersection of itself and another.'''
        itms = self.getdattr('synsattr', set()) 
        itms &= self.obj2set(other)
        self.setdattr('synsattr', itms)
        return self    
    def __ior__(self, other):
        '''Update a set with the union of itself and another.'''
        itms = self.getdattr('synsattr', set()) 
        itms |= self.obj2set(other)
        self.setdattr('synsattr', itms)
        return self
    def __isub__(self, other):
        '''Remove all elements of another set from this set.'''
        itms = self.getdattr('synsattr', set()) 
        itms -= self.obj2set(other)
        self.setdattr('synsattr', itms)
        return self
    def __ixor__(self, other):
        '''Update a set with the symmetric difference of itself and another.'''
        itms = self.getdattr('synsattr', set()) 
        itms ^= self.obj2set(other)
        self.setdattr('synsattr', itms)
        return self
    def __le__(self, other):
        '''Check if one set is a subset of another.'''
        return self.known <= self.obj2set(other)
    def __lt__(self, other):
        '''Check if one set is a proper subset of another.'''
        return self.known < self.obj2set(other)
    def __or__(self, other):
        '''Return the union of sets as a new set.'''
        return self.known | self.obj2set(other)
    def __rand__(self, other):
        '''Return the intersection of two sets as a new set.'''
        return self.known & self.obj2set(other)
    def __ror__(self, other):
        '''Return the union of sets as a new set.'''
        return self.known | self.obj2set(other)
    def __xor__(self, other):
        '''Return the symmetric difference of two sets as a new set.'''
        return self.known ^ self.obj2set(other)
    def __sub__(self, other):
        '''Return the difference of two sets as a new set.'''
        return self.known - self.obj2set(other)
    def __rsub__(self, other):
        '''Return the difference of two or more sets as a new set.'''
        return self.obj2set(other) - self.known
    def __rxor__(self, other):
        '''Return the symmetric difference of two sets as a new set.'''
        return self.obj2set(other) ^ self.known
    def add(self, syn: str):
        '''Add a synonym to the set'''
        self.known.add(syn)
    def remove(self, syn: str):
        '''Remove a synonym from the set'''
        self.known.remove(syn)
    def clear(self):
        '''Remove all synonyms from the set'''
        self.known.clear()
    def difference(self, *others):
        '''Return the difference of two or more sets as a new set.'''
        return self.known.difference(*[self.obj2set(o) for o in others])
    def difference_update(self, *others):
        '''Remove all elements of another set from this set.'''
        self.known.difference_update(*[self.obj2set(o) for o in others])
    def discard(self, syn: str):
        '''Remove a synonym from the set'''
        self.known.discard(syn)
    def intersection(self, *others):
        '''Return the intersection of two or more sets as a new set.'''
        return self.known.intersection(*[self.obj2set(o) for o in others])
    def intersection_update(self, *others):
        '''Update a set with the intersection of itself and another.'''
        self.known.intersection_update(*[self.obj2set(o) for o in others])
    def isdisjoint(self, other):
        '''Return True if two sets have a null intersection.'''
        return self.known.isdisjoint(self.obj2set(other))
    def issubset(self, other):
        '''Report whether another set contains this set.'''
        return self.known.issubset(self.obj2set(other))
    def issuperset(self, other):
        '''Report whether this set contains another set.'''
        return self.known.issuperset(self.obj2set(other))
    def pop(self):
        '''Remove and return an arbitrary set element.'''
        return self.known.pop()
    def remove(self, syn: str):
        '''Remove a synonym from the set'''
        self.known.remove(syn)
    def symmetric_difference(self, other):
        '''Return the symmetric difference of two sets as a new set.'''
        return self.known.symmetric_difference(self.obj2set(other))
    def symmetric_difference_update(self, other):
        '''Update a set with the symmetric difference of itself and another.'''
        self.known.symmetric_difference_update(self.obj2set(other))
    def union(self, *others):
        '''Return the union of sets as a new set.'''
        return self.known.union(*[self.obj2set(o) for o in others])
    def update(self, *others):
        '''Update a set with the union of itself and others.'''
        self.known.update(*[self.obj2set(o) for o in others])
